from asyncio import gather

import click

from ..api import GithubApi
from .utils import get_repos
from ..utils import wait


@click.command()
@click.pass_context
@click.option(
    '--list', is_flag=True,
    help='list open milestones', default=False)
@click.option('--close', help='milestone to close', default='')
def milestones(ctx, list, close):
    """View/edit/close milestones on github
    """
    repos = get_repos(ctx.parent.agile.get('labels'))
    if list:
        wait(_list_milestones(repos))
    elif close:
        click.echo('Closing milestones "%s"' % close)
        wait(_close_milestone(repos, close))
    else:
        click.echo(ctx.get_help())


async def _list_milestones(repos):
    git = GithubApi()
    requests = []
    for repo in repos:
        repo = git.repo(repo)
        requests.append(repo.milestones.get_list())
    data = await gather(*requests)
    milestones = set()
    for repo in data:
        milestones.update((data['title'] for data in repo))
    for title in sorted(milestones):
        click.echo(title)


def _close_milestone(repos, milestone):
    git = GithubApi()
    requests = []
    for repo in repos:
        repo = git.repo(repo)
        requests.append(_close_repo_milestone(repo, milestone))
    return gather(*requests)


async def _close_repo_milestone(repo, milestone):
    milestones = await repo.milestones.get_list()
    for m in milestones:
        if m['title'] == milestone:
            await repo.milestones.update(m, {'state': 'closed'})
            click.echo('Closed milestone %s' % m['html_url'])



from .components import Commits, Pulls, Issues, Component, Milestones
from .releases import Releases


class GitRepo(Component):
    """Github repository endpoints
    """
    def __init__(self, client, repo_path):
        super().__init__(client)
        self.repo_path = repo_path
        self.releases = Releases(self)
        self.commits = Commits(self)
        self.issues = Issues(self)
        self.pulls = Pulls(self)
        self.milestones = Milestones(self)
        self.headers = dict(
            accept='application/vnd.github.symmetra-preview+json'
        )

    @property
    def api_url(self):
        return '%s/repos/%s' % (self.client, self.repo_path)

    def label(self, name, color, update=True):
        """Create or update a label
        """
        url = '%s/labels' % self
        data = dict(name=name, color=color)
        response = self.http.post(
            url, json=data, auth=self.auth, headers=self.headers
        )
        if response.status_code == 201:
            return True
        elif response.status_code == 422 and update:
            url = '%s/%s' % (url, name)
            response = self.http.patch(
                url, json=data, auth=self.auth, headers=self.headers
            )
        response.raise_for_status()
        return False



import os
from slackclient import SlackClient


class SlackIntegration:
    token = os.environ.get('SLACK_TOKEN')

    def __init__(self, manager):
        self.manager = manager
        self.channel = manager.config.get('slack_channel', 'agile')
        self.cli = SlackClient(self.token)

    @classmethod
    def add(cls, manager):
        if cls.token:
            manager.message_brokers.append(cls(manager))

    def __call__(self, message):
        info = self.manager.info
        repo = self.manager.github_repo().repo_path
        sha = info['head']['id']
        commit = 'https://github.com/%s/commit/%s' % (repo, sha)
        message = '*%s* - %s\nby %s <%s>\n%s' % (
            repo,
            info['branch'],
            info['head']['committer_name'],
            info['head']['committer_email'],
            message
        )
        attachments = [
            dict(
                fallback="Git commit %s" % commit,
                actions=[
                    dict(
                        type="button",
                        text="Git commit",
                        url=commit
                    )
                ]
            )
        ]

        self.cli.api_call(
            'chat.postMessage',
            channel=self.channel,
            text=message,
            attachments=attachments
        )



from click.testing import CliRunner

from agiletoolkit.commands import start


def test_git():
    runner = CliRunner()
    result = runner.invoke(start, ['kong', '--help'])
    assert result.exit_code == 0
    assert result.output.startswith('Usage:')



import json

import click

from ..utils import gitrepo


@click.command()
def info():
    """Display information about repository
    """
    info = gitrepo()
    click.echo(json.dumps(info, indent=4))



from click.testing import CliRunner

from agiletoolkit.commands import start
from agiletoolkit.repo import RepoManager
from agiletoolkit.test import gitrepo
from agiletoolkit.api import GithubException


def mocked(*a, **kw):
    print('mocked')


def test_main():
    runner = CliRunner()
    result = runner.invoke(start)
    assert result.exit_code == 0
    assert result.output.startswith('Usage:')
    #
    result = runner.invoke(start, '--config', 'tests/cfg1.json')
    assert result.exit_code == 2
    assert result.output.startswith('Usage:')


def test_repo(mocker):
    with gitrepo('deploy'):
        m = RepoManager()
        m.validate_version = mocker.Mock(return_value=True)
        assert m.can_release()

    with gitrepo('deploy'):
        m = RepoManager()
        m.validate_version = mocker.Mock(side_effect=GithubException('test'))
        assert m.can_release() is False


def test_docker_tag():
    with gitrepo('feature/ABC-123/bla-bla', head_id='1234567890'):
        m = RepoManager()
        assert m.version() == 'feature-abc-123-bla-bla-12345678'



from unittest import mock
from contextlib import contextmanager

from . import utils
from .api import GithubException


@contextmanager
def gitrepo(branch, pr=False, tag=None, head_id=None):

    original_gitrepo = utils.gitrepo

    def mocker(root=None):
        data = original_gitrepo(root)
        data['branch'] = branch
        data['pr'] = pr
        data['tag'] = tag
        if head_id:
            data['head']['id'] = head_id
        return data

    with mock.patch('agiletoolkit.utils.gitrepo', side_effect=mocker) as m:
        yield m


def github_error(*args, **kwargs):
    raise GithubException



import click

from ..repo import RepoManager


@click.command()
@click.pass_context
def validate(ctx):
    """Check if version of repository is semantic
    """
    m = RepoManager(ctx.obj['agile'])
    click.echo(m.validate_version())



import click

from ..utils import command, niceJson
from ..repo import RepoManager


@click.command()
@click.pass_context
@click.option(
    '--yes', is_flag=True,
    help='Commit changes to github', default=False)
def release(ctx, yes):
    """Create a new release in github
    """
    with command():
        m = RepoManager(ctx.obj['agile'])
        branch = m.info['branch']
        if m.can_release('stage'):
            version = m.validate_version()
            name = 'v%s' % version
            body = ['Release %s from agiletoolkit' % name]
            api = m.github_repo()
            data = dict(
                tag_name=name,
                target_commitish=branch,
                name=name,
                body='\n\n'.join(body),
                draft=False,
                prerelease=False
            )
            if yes:
                data = m.wait(api.releases.create(data=data))
                m.message('Successfully created a new Github release')
            click.echo(niceJson(data))
        else:
            click.echo('skipped')



import json
from unittest import mock

from click.testing import CliRunner

from agiletoolkit import __version__
from agiletoolkit.commands import start
from agiletoolkit.test import gitrepo

from .utils import mock_response


def test_git():
    runner = CliRunner()
    result = runner.invoke(start, ['git'])
    assert result.exit_code == 0
    assert result.output.startswith('Usage:')


def test_git_validate():
    runner = CliRunner()

    with gitrepo('deploy'):
        result = runner.invoke(start, ['git', 'validate'])
        assert result.exit_code == 0
        assert result.output.strip() == __version__


def test_git_info():
    runner = CliRunner()
    result = runner.invoke(start, ['git', 'info'])
    assert result.exit_code == 0
    data = json.loads(result.output.strip())
    assert 'branch' in data


def test_git_remote():
    runner = CliRunner()
    result = runner.invoke(start, ['git', 'remote'])
    assert result.exit_code == 0
    assert result.output.strip() == 'lendingblock/agile-toolkit'


def __test_git_release():
    runner = CliRunner()

    with gitrepo('deploy') as mock:
        result = runner.invoke(start, ['git', 'release'])
        assert result.exit_code == 0
        assert mock.called
        data = json.loads(result.output.strip())
        assert data['name'] == 'v%s' % __version__


def test_git_release_skipped():
    runner = CliRunner()

    with gitrepo('master') as mock:
        result = runner.invoke(start, ['git', 'release'])
        assert result.exit_code == 0
        assert mock.called
        assert result.output.strip() == 'skipped'


def test_labels_error():
    runner = CliRunner()
    result = runner.invoke(start, [
        '--config', 'tests/cfg1.json', 'git', 'labels'
    ])
    assert result.exit_code == 1
    assert result.output == (
        'Error: You need to specify the "repos" list in the config\n'
    )
    result = runner.invoke(start, [
        '--config', 'tests/cfg2.json', 'git', 'labels'
    ])
    assert result.exit_code == 1
    assert result.output == (
        'Error: You need to specify the "labels" dictionary in the config\n'
    )


def test_git_labels():
    runner = CliRunner()
    with mock.patch('requests.Session') as ses:
        ses.return_value = mock_response()
        result = runner.invoke(start, [
            '--config', 'tests/cfg3.json', 'git', 'labels'
        ])
        assert result.exit_code == 0
